/*
 *  docpdfy by Davide Libenzi (MS Word DOC to Adobe PDF converter)
 *  Copyright (C) 2006  Davide Libenzi
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Davide Libenzi <davidel@xmailserver.org>
 *
 */

#include "stdafx.h"
#include "CApplication.h"
#include "CDocument0.h"
#include "CDocuments.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif


// The one and only application object
CWinApp theApp;

using namespace std;

static COleVariant cvOptional((long) DISP_E_PARAMNOTFOUND, VT_ERROR),
                   cvFalse((SHORT) FALSE), cvTrue((SHORT) TRUE);


void ErrorPrint(TCHAR const *pszFmt, ...)
{
    LPVOID lpMsgBuf;
    va_list Args;

    FormatMessage( 
        FORMAT_MESSAGE_ALLOCATE_BUFFER | 
        FORMAT_MESSAGE_FROM_SYSTEM | 
        FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL,
        GetLastError(),
        0,
        (LPTSTR) &lpMsgBuf,
        0,
        NULL 
        );
    va_start(Args, pszFmt);
    _vftprintf(stderr, pszFmt, Args);
    va_end(Args);
    _ftprintf(stderr, _T("%s\n"), lpMsgBuf);
    LocalFree(lpMsgBuf);
}

TCHAR *GetRootedPath(TCHAR const *pszPath)
{
    size_t iLen, iDLen;
    TCHAR *pszRPath;
    TCHAR szCDir[MAX_PATH];

    if (*pszPath != TCHAR('.'))
        return _tcsdup(pszPath);
    GetCurrentDirectory(MAX_PATH, szCDir);
    iLen = (iDLen = _tcslen(szCDir)) + _tcslen(pszPath) + 3;
    if ((pszRPath = (TCHAR *) malloc(iLen * sizeof(TCHAR))) == NULL) {
        ErrorPrint(_T("Unable to allocate memory\n"));
        return NULL;
    }
    _tcscpy_s(pszRPath, iLen, szCDir);
    if (iDLen > 0 && pszRPath[iDLen - 1] != TCHAR('\\'))
        pszRPath[iDLen++] = TCHAR('\\');
    _tcscpy_s(pszRPath + iDLen, iLen - iDLen, pszPath);

    return pszRPath;
}

TCHAR *FileExtChange(TCHAR const *pszFilePath, TCHAR const *pszExt)
{
    size_t iLen = _tcslen(pszFilePath) + _tcslen(pszExt) + 2;
    TCHAR *pszPath, *pszDot;

    if ((pszPath = (TCHAR *) malloc(iLen * sizeof(TCHAR))) == NULL) {
        ErrorPrint(_T("Unable to allocate memory\n"));
        return NULL;
    }
    _tcscpy_s(pszPath, iLen, pszFilePath);
    if ((pszDot = _tcsrchr(pszPath, TCHAR('.'))) == NULL)
        pszDot = pszPath + _tcslen(pszPath);
    _tcscpy_s(pszDot, iLen - _tcslen(pszPath), pszExt);

    return pszPath;
}

TCHAR *FileExtRebase(TCHAR const *pszFilePath, TCHAR const *pszExt, TCHAR const *pszOutDir)
{
    size_t iLen, iDLen;
    TCHAR *pszPath, *pszDot;
    TCHAR const *pszSlash;

    if (pszOutDir == NULL)
        return FileExtChange(pszFilePath, pszExt);
    if ((pszSlash = _tcsrchr(pszFilePath, TCHAR('\\'))) != NULL)
        pszFilePath = pszSlash + 1;
    iLen = (iDLen = _tcslen(pszOutDir)) + _tcslen(pszFilePath) + _tcslen(pszExt) + 2;
    if ((pszPath = (TCHAR *) malloc(iLen * sizeof(TCHAR))) == NULL) {
        ErrorPrint(_T("Unable to allocate memory\n"));
        return NULL;
    }
    _tcscpy_s(pszPath, iLen, pszOutDir);
    if (iDLen > 0 && pszPath[iDLen - 1] != TCHAR('\\'))
        pszPath[iDLen++] = TCHAR('\\');
    _tcscpy_s(pszPath + iDLen, iLen - iDLen, pszFilePath);
    if ((pszDot = _tcsrchr(pszPath, TCHAR('.'))) == NULL)
        pszDot = pszPath + _tcslen(pszPath);
    _tcscpy_s(pszDot, iLen - _tcslen(pszPath), pszExt);

    return pszPath;
}

int Doc2Prn(CApplication *pApp, CDocuments *pDocs, TCHAR const *pszDocFile, TCHAR const *pszPrnFile)
{
    LPDISPATCH lpDoc;

    if ((lpDoc = pDocs->Open(
        COleVariant((BSTR) pszDocFile),
        cvFalse,    // Confirm Conversion.
        cvTrue,    // ReadOnly.
        cvFalse,    // AddToRecentFiles.
        cvOptional, // PasswordDocument.
        cvOptional, // PasswordTemplate.
        cvFalse,    // Revert.
        cvOptional, // WritePasswordDocument.
        cvOptional, // WritePasswordTemplate.
        cvOptional, // Format.
        cvOptional, // Encoding // New for Word 2000/2002
        cvTrue,     // Visible
        cvOptional, // OpenConflictDocument
        cvOptional, // OpenAndRepair
        COleVariant((long) 0),     // DocumentDirection wdDocumentDirection LeftToRight
        cvOptional  // NoEncodingDialog
        )) == NULL) {
            ErrorPrint(_T("Unable to open document: %s\n"), pszDocFile);
            return -1;
    }

    TCHAR *pszPrnPath = NULL;

    if (pszPrnFile == NULL) {
        if ((pszPrnPath = FileExtChange(pszDocFile, _T(".prn"))) == NULL) {
            lpDoc->Release();
            return -1;
        }
        pszPrnFile = pszPrnPath;
    }   

    CDocument0 _doc(lpDoc);

    _doc.PrintOut(
        cvFalse,    // Background
        cvOptional, // Append
        cvOptional, // Range
        COleVariant((BSTR) pszPrnFile), // OutputFileName
        cvOptional, // From
        cvOptional, // To
        cvOptional, // Item
        cvOptional, // Copies
        cvOptional, // Pages
        cvOptional, // PageType
        cvTrue, // PrintToFile
        cvOptional, // Collate
        cvOptional, // ActivePrinterMacGX
        cvOptional, // ManualDuplexPrint
        cvOptional, // PrintZoomColumn
        cvOptional, // PrintZoomRow
        cvOptional, // PrintZoomPaperWidth
        cvOptional  // PrintZoomPaperHeight
        );
    free(pszPrnPath);
    while (pApp->get_BackgroundPrintingStatus() > 0)
        Sleep(1000);
    _doc.Close(cvFalse, cvOptional, cvOptional);

    return 0;
}

int Prn2Pdf(TCHAR const *pszPrnFile, TCHAR const *pszPdfFile)
{
    TCHAR *pszPdfPath = NULL, *pszCmdLine;

    if (pszPdfFile == NULL) {
        if ((pszPdfPath = FileExtChange(pszPrnFile, _T(".pdf"))) == NULL)
            return -1;
        pszPdfFile = pszPdfPath;
    }

    size_t iCmdLnSize = _tcslen(pszPrnFile) + _tcslen(pszPdfFile) + 256;

    if ((pszCmdLine = (TCHAR *) malloc(iCmdLnSize * sizeof(TCHAR))) == NULL) {
        ErrorPrint(_T("Unable to allocate memory\n"));
        return -1;
    }
    _sntprintf_s(pszCmdLine, iCmdLnSize, iCmdLnSize,
        _T("gswin32c.exe -q -dNOPAUSE -dBATCH -sDEVICE=pdfwrite -sOutputFile=\"%s\" -c save pop -f \"%s\""),
        pszPdfFile, pszPrnFile);
    free(pszPdfPath);

    STARTUPINFO SInf;
    PROCESS_INFORMATION PInf;

    ZeroMemory(&SInf, sizeof(SInf));
    SInf.cb = sizeof(SInf);
    ZeroMemory(&PInf, sizeof(PInf));
    if (!CreateProcess(NULL,
        pszCmdLine,
        NULL,
        NULL,
        FALSE,
        0,
        NULL,
        NULL,
        &SInf,
        &PInf)) {
        ErrorPrint(_T("Unable to execute process: %s\n"), pszCmdLine);
        free(pszCmdLine);
        return -1;
    }
    free(pszCmdLine);
    WaitForSingleObject(PInf.hProcess, INFINITE);
    CloseHandle(PInf.hProcess);
    CloseHandle(PInf.hThread);

    return 0;
}

int Doc2Pdf(CApplication *pApp, CDocuments *pDocs, TCHAR const *pszDocFile, TCHAR const *pszOutDir)
{
    int iResult;
    TCHAR *pszPrnFile, *pszPdfFile;

    pszPrnFile = FileExtRebase(pszDocFile, _T(".prn"), pszOutDir);
    pszPdfFile = FileExtRebase(pszDocFile, _T(".pdf"), pszOutDir);
    if (pszPrnFile == NULL || pszPdfFile == NULL ||
        Doc2Prn(pApp, pDocs, pszDocFile, pszPrnFile) < 0) {
        free(pszPdfFile);
        free(pszPrnFile);
        return -1;
    }
    
    iResult = Prn2Pdf(pszPrnFile, pszPdfFile);

    DeleteFile(pszPrnFile);
    free(pszPdfFile);
    free(pszPrnFile);

    return iResult;
}

void Usage(TCHAR const *pszPrg)
{
    _ftprintf(stderr, _T("Use: %s [-D OUTDIR] [-p PRINTER] docfile ...\n"), pszPrg);
}

int _tmain(int argc, TCHAR* argv[], TCHAR* envp[])
{
    int i;
    TCHAR *pszOutDir = NULL;
    TCHAR const *pszPrinter = NULL;

    _ftprintf(stdout, _T("docpdfy - DOC to PDF batch conversion utility\n"));
    _ftprintf(stdout, _T("          Copyright (C) 2006  Davide Libenzi <davidel@xmailserver.org>\n\n"));
    if (!AfxWinInit(::GetModuleHandle(NULL), NULL, ::GetCommandLine(), 0)) {
        ErrorPrint(_T("MFC initialization failed\n"));
        return 1;
    }
    CoInitialize(NULL);
    for (i = 1; i < argc; i++) {
        if (_tcscmp(argv[i], _T("-D")) == 0) {
            if (++i < argc)
                pszOutDir = GetRootedPath(argv[i]);
        } else if (_tcscmp(argv[i], _T("-p")) == 0) {
            if (++i < argc)
                pszPrinter = argv[i];
        } else
            break;
    }
    if (i == argc) {
        Usage(argv[0]);
        return 1;
    }

    CApplication _app;

    if (!_app.CreateDispatch(_T("Word.Application"))) {
        ErrorPrint(_T("Unable to get 'Word.Application' interface\n"));
        return 1;
    }

    LPDISPATCH lpDocs;

    if ((lpDocs = _app.get_Documents()) == NULL) {
        ErrorPrint(_T("Unable to get Documents interface\n"));
        return 2;
    }

    CDocuments _docs(lpDocs);
    CString sOldPrinter;

    if (pszPrinter != NULL) {
        sOldPrinter = _app.get_ActivePrinter();
        _app.put_ActivePrinter(pszPrinter);
    }

    for (; i < argc; i++) {
        TCHAR *pszDocFile = GetRootedPath(argv[i]);

        _ftprintf(stdout, _T("\nProcessing: %s\n"), argv[i]);
        Doc2Pdf(&_app, &_docs, pszDocFile, pszOutDir);
        free(pszDocFile);
    }
    free(pszOutDir);

    if (_docs.get_Count() > 0)
        _docs.Close(cvFalse, cvOptional, cvOptional);
    if (pszPrinter != NULL)
        _app.put_ActivePrinter(sOldPrinter);
    _app.Quit(cvFalse, cvOptional, cvOptional);

    return 0;
}

